@@ -2,6 +2,7 @@
 """
 Supports the definition of commands in separate classes to be composed into cmd2.Cmd
 """
+
 from typing import (
     TYPE_CHECKING,
     Callable,
@@ -9,6 +10,7 @@ from typing import (
     Mapping,
     Optional,
     Type,
+    TypeVar,
 )
 
 from .constants import (
@@ -29,8 +31,10 @@ if TYPE_CHECKING:  # pragma: no cover
 #: Further refinements are needed to define the input parameters
 CommandFunc = Callable[..., Optional[bool]]
 
+CommandSetType = TypeVar('CommandSetType', bound=Type['CommandSet'])
 
-def with_default_category(category: str, *, heritable: bool = True) -> Callable[[Type['CommandSet']], Type['CommandSet']]:
+
+def with_default_category(category: str, *, heritable: bool = True) -> Callable[[CommandSetType], CommandSetType]:
     """
     Decorator that applies a category to all ``do_*`` command methods in a class that do not already
     have a category specified.
@@ -41,7 +45,7 @@ def with_default_category(category: str, *, heritable: bool = True) -> Callable[
     override the default category.
 
     If `heritable` is set to False, then only the commands declared locally to this CommandSet will be placed in the
-    specified category. Dynamically created commands, and commands declared in sub-classes will not receive this
+    specified category. Dynamically created commands and commands declared in sub-classes will not receive this
     category.
 
     :param category: category to put all uncategorized commands in
@@ -49,7 +53,7 @@ def with_default_category(category: str, *, heritable: bool = True) -> Callable[
     :return: decorator function
     """
 
-    def decorate_class(cls: Type[CommandSet]) -> Type[CommandSet]:
+    def decorate_class(cls: CommandSetType) -> CommandSetType:
         if heritable:
             setattr(cls, CLASS_ATTR_DEFAULT_HELP_CATEGORY, category)
 
@@ -91,20 +95,42 @@ class CommandSet(object):
     """
 
     def __init__(self) -> None:
-        self._cmd: Optional[cmd2.Cmd] = None
+        # Private reference to the CLI instance in which this CommandSet running.
+        # This will be set when the CommandSet is registered and it should be
+        # accessed by child classes using the self._cmd property.
+        self.__cmd_internal: Optional[cmd2.Cmd] = None
+
         self._settables: Dict[str, Settable] = {}
         self._settable_prefix = self.__class__.__name__
 
+    @property
+    def _cmd(self) -> 'cmd2.Cmd':
+        """
+        Property for child classes to access self.__cmd_internal.
+
+        Using this property ensures that self.__cmd_internal has been set
+        and it tells type checkers that it's no longer a None type.
+
+        Override this property if you need to change its return type to a
+        child class of Cmd.
+
+        :raises CommandSetRegistrationError: if CommandSet is not registered.
+        """
+        if self.__cmd_internal is None:
+            raise CommandSetRegistrationError('This CommandSet is not registered')
+        return self.__cmd_internal
+
     def on_register(self, cmd: 'cmd2.Cmd') -> None:
         """
         Called by cmd2.Cmd as the first step to registering a CommandSet. The commands defined in this class have
-        not be added to the CLI object at this point. Subclasses can override this to perform any initialization
+        not been added to the CLI object at this point. Subclasses can override this to perform any initialization
         requiring access to the Cmd object (e.g. configure commands and their parsers based on CLI state data).
 
         :param cmd: The cmd2 main application
+        :raises CommandSetRegistrationError: if CommandSet is already registered.
         """
-        if self._cmd is None:
-            self._cmd = cmd
+        if self.__cmd_internal is None:
+            self.__cmd_internal = cmd
         else:
             raise CommandSetRegistrationError('This CommandSet has already been registered')
 
@@ -128,7 +154,7 @@ class CommandSet(object):
         Called by ``cmd2.Cmd`` after a CommandSet has been unregistered and all its commands removed from the CLI.
         Subclasses can override this to perform remaining cleanup steps.
         """
-        self._cmd = None
+        self.__cmd_internal = None
 
     @property
     def settable_prefix(self) -> str:
@@ -144,7 +170,7 @@ class CommandSet(object):
 
         :param settable: Settable object being added
         """
-        if self._cmd:
+        if self.__cmd_internal is not None:
             if not self._cmd.always_prefix_settables:
                 if settable.name in self._cmd.settables.keys() and settable.name not in self._settables.keys():
                     raise KeyError(f'Duplicate settable: {settable.name}')
@@ -159,9 +185,18 @@ class CommandSet(object):
         Convenience method for removing a settable parameter from the CommandSet
 
         :param name: name of the settable being removed
-        :raises: KeyError if the Settable matches this name
+        :raises KeyError: if the Settable matches this name
         """
         try:
             del self._settables[name]
         except KeyError:
             raise KeyError(name + " is not a settable parameter")
+
+    def sigint_handler(self) -> bool:
+        """
+        Handle a SIGINT that occurred for a command in this CommandSet.
+
+        :return: True if this completes the interrupt handling and no KeyboardInterrupt will be raised.
+                 False to raise a KeyboardInterrupt.
+        """
+        return False
